﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace GameOfLife
{
    /// <summary>
    /// Stores the state of the grid, and also provides methods for updating
    /// the grid to the next round.
    /// </summary>
    internal class Grid
    {
        /// <summary>
        /// Stores the actual values. True means the cell is alive, false 
        /// means it is dead.
        /// </summary>
        private bool[,] values;

        /// <summary>
        /// Gets or sets the number of rows in the grid.
        /// </summary>
        public int Rows { get; private set; }

        /// <summary>
        /// Gets or sets the number of columns in the grid.
        /// </summary>
        public int Columns { get; private set; }

        /// <summary>
        /// Indicates if the grid should wrap (it's a toroidal wrap, where
        /// you can wrap from top to bottom or side to side). If this is true,
        /// then cells on the other side of the wrapping will be counted as
        /// neighbors. If this is false, they won't be.
        /// </summary>
        public bool Wrap { get; set; }

        /// <summary>
        /// Creates a new grid with the default size of 36x36.
        /// </summary>
        public Grid()
            : this(36, 36) // This is using constructor chaining.
        {
        }
        
        /// <summary>
        /// Creates a new grid with a given number of rows and columsn.
        /// </summary>
        /// <param name="rows"></param>
        /// <param name="columns"></param>
        public Grid(int rows, int columns)
        {
            values = new bool[rows, columns];
            Rows = rows;
            Columns = columns;
            Wrap = false;
        }

        /// <summary>
        /// Fills the grid randomly, based on a certain percentage. A percentage
        /// of 0 means the grid will be completely empty. A percentage of 1
        /// means that it will be completely full.
        /// </summary>
        /// <param name="percent"></param>
        public void Randomize(double percent)
        {
            Random random = new Random();

            for (int row = 0; row < Rows; row++)
            {
                for (int column = 0; column < Columns; column++)
                {
                    if (random.NextDouble() < percent)
                    {
                        values[row, column] = true;
                    }
                    else
                    {
                        values[row, column] = false;
                    }
                }
            }
        }

        /// <summary>
        /// Advances to the next "generation" or iteration of the game.
        /// </summary>
        public void Update()
        {
            // THIS IS VERY IMPORTANT:
            // You MUST work with a different set of data. Since neighbors
            // depend on the value of their neighbors, if you modify the
            // original, you may not end up with the correct total for the
            // number of neighbors it has. So we'll write to this new array,
            // and then replace the old array with this one when we're done.
            bool[,] nextGeneration = new bool[Rows, Columns];

            // Go through every cell in the grid.
            for (int row = 0; row < Rows; row++)
            {
                for (int column = 0; column < Columns; column++)
                {
                    // Figure out the important information we need to decide
                    // what happens next.
                    int count = CountNeighbors(row, column);
                    bool isCurrentlyAlive = values[row, column];
                    
                    // Execute our rules for cell birth and death.
                    if (isCurrentlyAlive && count > 3) 
                            { isCurrentlyAlive = false; }
                    else if (isCurrentlyAlive && count < 2) 
                            { isCurrentlyAlive = false; }
                    else if (!isCurrentlyAlive && count == 3) 
                            { isCurrentlyAlive = true; }

                    nextGeneration[row, column] = isCurrentlyAlive;
                }
            }

            values = nextGeneration;
        }

        /// <summary>
        /// Draws the current grid.
        /// </summary>
        public void Draw()
        {
            Console.Clear();

            // The border is drawn in yellow, empty cells are marked with
            // black, and live cells are blue.
            Console.ForegroundColor = ConsoleColor.Yellow;
            Console.Write("███");
            for (int column = 0; column < Columns + 1; column++)
            {
                Console.Write("██");
            }
            Console.WriteLine();


            for (int row = 0; row < Rows; row++)
            {
                Console.ForegroundColor = ConsoleColor.Yellow;
                Console.Write("██ ");

                for (int column = 0; column < Columns; column++)
                {
                    bool alive = values[row, column];

                    if (alive)
                    {
                        Console.ForegroundColor = ConsoleColor.Blue; 
                        Console.Write("O ");
                    }
                    else
                    {
                        Console.ForegroundColor = ConsoleColor.DarkGray; 
                        Console.Write(". ");
                    }
                }

                Console.ForegroundColor = ConsoleColor.Yellow;
                Console.Write("██");
                Console.WriteLine();
            }

            Console.ForegroundColor = ConsoleColor.Yellow;
            Console.Write("███");
            for (int column = 0; column < Columns + 1; column++)
            {
                Console.Write("██");
            }
            Console.WriteLine();
        }

        /// <summary>
        /// Counts the number of alive neighbors for a given cell.
        /// </summary>
        /// <param name="row"></param>
        /// <param name="column"></param>
        /// <returns></returns>
        public int CountNeighbors(int row, int column)
        {
            int count = 0;
            if (IsCellAlive(row - 1, column - 1)) { count++; }
            if (IsCellAlive(row - 1, column)) { count++; }
            if (IsCellAlive(row - 1, column + 1)) { count++; }
            if (IsCellAlive(row, column - 1)) { count++; }
            if (IsCellAlive(row, column + 1)) { count++; }
            if (IsCellAlive(row + 1, column - 1)) { count++; }
            if (IsCellAlive(row + 1, column)) { count++; }
            if (IsCellAlive(row + 1, column + 1)) { count++; }

            return count;
        }

        /// <summary>
        /// Determines if a single cell is alive.
        /// </summary>
        private bool IsCellAlive(int row, int column)
        {
            // This makes the decision about whether to use wrapped or
            // unwrapped neighbor counting, and delegates to the right
            // alternative method.
            if (Wrap) { return IsCellAliveWrapped(row, column); }
            else { return IsCellAliveUnwrapped(row, column); }
        }

        /// <summary>
        /// Determines if a cell is alive, assuming that the grid wraps around.
        /// </summary>
        /// <param name="row"></param>
        /// <param name="column"></param>
        /// <returns></returns>
        private bool IsCellAliveWrapped(int row, int column)
        {
            // If the row or column is negative, the remainder operator won't
            // work correctly for what we want to do here. So we add the total
            // rows/columns to make sure that the values are for sure positive.
            // Once we can be certain it is positive, we can simply use the
            // remainder operator to figure out where it actually is, even
            // if it wraps around.
            row = (row + Rows) % Rows;
            column = (column + Columns) % Columns;

            return values[row, column];
        }

        /// <summary>
        /// Determines if a cell is alive, assuming that the board does not
        /// wrap around. If row or column are out of the valid range, false
        /// will be returned.
        /// </summary>
        /// <param name="row"></param>
        /// <param name="column"></param>
        /// <returns></returns>
        private bool IsCellAliveUnwrapped(int row, int column)
        {
            // Check the border conditions.
            if (row < 0) { return false; }
            if (row >= Rows) { return false; }
            if (column < 0) { return false; }
            if (column >= Columns) { return false; }

            return values[row, column];
        }
    }
}
